# -*- coding: utf-8 -*-
#
# This file is part of Flask-WebpackExt
# Copyright (C) 2017 CERN.
#
# Flask-WebpackExt is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Minimal Flask application example.

First install Flask-WebpackExt, create the instance folder and export
environment variables:

.. code-block:: console

   $ pip install -e .[all]
   $ cd examples
   $ mkdir instance
   $ export FLASK_APP=app.py FLASK_ENV=development

Next, install and build the assets with webpack:

.. code-block:: console

   $ flask webpack buildall

Start the development server:

.. code-block:: console

   $ flask run

and open the example application in your browser:

.. code-block:: console

    $ open http://127.0.0.1:5000/

"""

from __future__ import absolute_import, print_function

from flask import Flask, render_template

from flask_webpackext import FlaskWebpackExt, WebpackTemplateProject

# Create a Webpack project.
project = WebpackTemplateProject(
    __name__,
    project_folder='assets',
    config_path='build/config.json',
)

# Create Flask application
app = Flask(
    __name__,
    template_folder='templates',
    static_folder='instance/static'
)
# Configure application
app.config.update(dict(
    WEBPACKEXT_PROJECT=project,
))

# Initialize extension
FlaskWebpackExt(app)


# Basic view
@app.route('/')
def index():
    """Frontpage."""
    return render_template('index.html')



# -*- coding: utf-8 -*-
#
# This file is part of Flask-WebpackExt
# Copyright (C) 2017 CERN.
#
# Flask-WebpackExt is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Proxy to current extension."""

from __future__ import absolute_import, print_function

from flask import current_app
from markupsafe import Markup
from pywebpack import Manifest, ManifestEntry, ManifestLoader

from .errors import ManifestKeyNotFoundError


class JinjaManifest(Manifest):
    """Manifest entry which marks rendered strings as safe for Jinja."""

    def __getitem__(self, key):
        """Get a manifest entry."""
        try:
            return super(JinjaManifest, self).__getitem__(key)
        except KeyError:
            raise ManifestKeyNotFoundError('Key {} not found in manifest.json'
                                           .format(key))

    def __getattr__(self, name):
        """Get a manifest entry."""
        try:
            return super(JinjaManifest, self).__getattr__(name)
        except AttributeError as e:
            raise ManifestKeyNotFoundError('Key {} not found in manifest.json'
                                           .format(name))


class JinjaManifestEntry(ManifestEntry):
    """Manifest entry which marks rendered strings as safe for Jinja."""

    def __html__(self):
        """Ensures that string is not escaped when included in Jinja."""
        return Markup(self.render())


class JinjaManifestLoader(ManifestLoader):
    """Factory which uses the Jinja manifest entry."""

    cache = {}

    def __init__(self, manifest_cls=JinjaManifest,
                 entry_cls=JinjaManifestEntry):
        """Initialize manifest loader."""
        super(JinjaManifestLoader, self).__init__(
            manifest_cls=manifest_cls,
            entry_cls=entry_cls
        )

    def load(self, filepath):
        """Load a manifest from a file."""
        if current_app.debug or filepath not in JinjaManifestLoader.cache:
            JinjaManifestLoader.cache[filepath] = \
                super(JinjaManifestLoader, self).load(filepath)
        return JinjaManifestLoader.cache[filepath]



# -*- coding: utf-8 -*-
#
# This file is part of Flask-WebpackExt
# Copyright (C) 2017 CERN.
#
# Flask-WebpackExt is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Webpack integration for Flask."""

from __future__ import absolute_import, print_function

from os.path import join

from pywebpack import FileStorage, ManifestLoader
from werkzeug.utils import import_string

from . import config
from ._compat import string_types
from .manifest import JinjaManifestLoader
from .proxies import current_manifest


class FlaskWebpackExt(object):
    """Flask-WebpackExt extension."""

    def __init__(self, app=None):
        """Extension initialization."""
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Flask application initialization."""
        self.init_config(app)
        app.add_template_global(current_manifest, name='webpack')
        app.extensions['flask-webpackext'] = _FlaskWebpackExtState(app)

    def init_config(self, app):
        """Initialize configuration."""
        app.config.setdefault(
            'WEBPACKEXT_PROJECT_DISTDIR',
            join(app.static_folder, 'dist'),
        )
        app.config.setdefault(
            'WEBPACKEXT_PROJECT_DISTURL',
            join(app.static_url_path, 'dist'),
        )
        app.config.setdefault(
            'WEBPACKEXT_PROJECT_BUILDDIR',
            join(app.instance_path, 'assets'),
        )
        app.config.setdefault(
            'WEBPACKEXT_STORAGE_CLS',
            FileStorage
        )
        app.config.setdefault(
            'WEBPACKEXT_MANIFEST_LOADER',
            JinjaManifestLoader,
        )

        for k in dir(config):
            if k.startswith('WEBPACKEXT_'):
                app.config.setdefault(k, getattr(config, k))


class _FlaskWebpackExtState(object):
    """Flask webpack state object."""

    def __init__(self, app):
        """Initialize state."""
        self.app = app

    @property
    def manifest_loader(self):
        """Manifest loader."""
        loader = self.app.config['WEBPACKEXT_MANIFEST_LOADER']
        if isinstance(loader, string_types):
            return import_string(loader)
        return loader

    @property
    def manifest(self):
        """Manifest."""
        path = self.app.config['WEBPACKEXT_MANIFEST_PATH']
        if path:
            return self.manifest_loader().load(
                join(self.app.static_folder, path))
        return None

    @property
    def project(self):
        """Webpack project."""
        project = self.app.config['WEBPACKEXT_PROJECT']
        if isinstance(project, string_types):
            return import_string(project)
        return project

    @property
    def storage_cls(self):
        """Default storage class."""
        cls_ = self.app.config['WEBPACKEXT_STORAGE_CLS']
        if isinstance(cls_, string_types):
            return import_string(cls_)
        return cls_



# -*- coding: utf-8 -*-
#
# This file is part of Flask-WebpackExt
# Copyright (C) 2017, 2018 CERN.
#
# Flask-WebpackExt is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Webpack project utilities for Flask-WebpackExt."""

from __future__ import absolute_import, print_function

from os.path import join

from flask import current_app
from flask.helpers import get_root_path
from pywebpack import WebpackBundle as PyWebpackBundle
from pywebpack import WebpackBundleProject as PyWebpackBundleProject
from pywebpack import WebpackTemplateProject as PyWebpackTemplateProject

from .proxies import current_webpack


def flask_config():
    """Flask configuration injected in Webpack.

    :return: Dictionary which contains the information Flask-WebpackExt knows
        about a Webpack project and the absolute URLs for static files and
        assets. The dictionary consists of a key ``build`` with the following
        keys inside:

        * ``debug``: Boolean which represents if Flask debug is on.
        * ``context``: Absolute path to the generated assets directory.
        * ``assetsPath``: Absolute path to the generated static directory.
        * ``assetsURL``: URL to access the built files.
        * ``staticPath``: Absolute path to the generated static directory.
        * ``staticURL``: URL to access the static files..
    """
    assets_url = current_app.config['WEBPACKEXT_PROJECT_DISTURL']
    if not assets_url.endswith('/'):
        assets_url += '/'
    static_url = current_app.static_url_path
    if not static_url.endswith('/'):
        static_url += '/'

    return {
        'build': {
            'debug': current_app.debug,
            'context': current_webpack.project.path,
            'assetsPath': current_app.config['WEBPACKEXT_PROJECT_DISTDIR'],
            'assetsURL': assets_url,
            'staticPath': current_app.static_folder,
            'staticURL': static_url,
        }
    }


class _PathStorageMixin(object):
    """Mixin class."""

    @property
    def path(self):
        """Get path to project."""
        return current_app.config['WEBPACKEXT_PROJECT_BUILDDIR']

    @property
    def storage_cls(self):
        """Get storage class."""
        return current_webpack.storage_cls


class WebpackTemplateProject(_PathStorageMixin, PyWebpackTemplateProject):
    """Flask webpack template project."""

    def __init__(self, import_name, project_folder=None, config=None,
                 config_path=None):
        """Initialize project.

        :param import_name: Name of the module where the
            WebpackTemplateProject class is instantiated. It is used to
            determine the absolute path to the ``project_folder``.
        :param project_folder: Relative path to the Webpack project.
        :param config: Dictionary which overrides the ``config.json`` file
            generated by Flask-WebpackExt. Use carefuly and only if you know
            what you are doing since ``config.json`` is the file that holds the
            key information to integrate Flask with Webpack.
        :param config_path: Path where Flask-WebpackExt is going to write the
            ``config.json``, this file is generated by
            :func:`flask_webpackext.project.flask_config`.
        """
        project_template_dir = join(get_root_path(import_name), project_folder)
        super(WebpackTemplateProject, self).__init__(
            None,
            project_template_dir=project_template_dir,
            config=config or flask_config,
            config_path=config_path,
        )


class WebpackBundleProject(_PathStorageMixin, PyWebpackBundleProject):
    """Flask webpack bundle project."""

    def __init__(self, import_name, project_folder=None, bundles=None,
                 config=None, config_path=None):
        """Initialize templated folder.

        :param import_name: Name of the module where the WebpackBundleProject
            class is instantiated. It is used to determine the absolute path
            to the ``project_folder``.
        :param project_folder: Relative path to the Webpack project which is
            going to aggregate all the ``bundles``.
        :param bundles: List of
            :class:`flask_webpackext.bundle.WebpackBundle`. This list can be
            statically defined if the bundles are known before hand, or
            dinamically generated using
            :func:`pywebpack.helpers.bundles_from_entry_point` so the bundles
            are discovered from the defined Webpack entrypoints exposed by
            other modules.
        :param config: Dictionary which overrides the ``config.json`` file
            generated by Flask-WebpackExt. Use carefuly and only if you know
            what you are doing since ``config.json`` is the file that holds the
            key information to integrate Flask with Webpack.
        :param config_path: Path where Flask-WebpackExt is going to write the
            ``config.json``, this file is generated by
            :func:`flask_webpackext.project.flask_config`.
        """
        project_template_dir = join(get_root_path(import_name), project_folder)
        super(WebpackBundleProject, self).__init__(
            None,
            project_template_dir=project_template_dir,
            bundles=bundles,
            config=config or flask_config,
            config_path=config_path,
        )



# -*- coding: utf-8 -*-
#
# This file is part of Flask-WebpackExt
# Copyright (C) 2017 CERN.
#
# Flask-WebpackExt is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Pytest configuration."""

from __future__ import absolute_import, print_function

import shutil
import tempfile
from os import makedirs
from os.path import dirname, join

import pytest
from click.testing import CliRunner
from flask import Flask
from flask.cli import ScriptInfo

from flask_webpackext import FlaskWebpackExt, WebpackBundle, \
    WebpackBundleProject, WebpackProject, WebpackTemplateProject


@pytest.yield_fixture(scope='function')
def tmpdir():
    """Temporary directory."""
    path = tempfile.mkdtemp()
    yield path
    shutil.rmtree(path)


@pytest.fixture(scope='function')
def instance_path(tmpdir):
    """Temporary instance path."""
    f = join(tmpdir, 'instance')
    makedirs(f)
    return f


@pytest.fixture(scope='function')
def static_folder(instance_path):
    """Static folder."""
    f = join(instance_path, 'static')
    makedirs(f)
    return f


@pytest.fixture(scope='function')
def manifest(static_folder):
    """Static folder."""
    src = join(dirname(__file__), 'manifest.json')
    dst = join(static_folder, 'dist/manifest.json')
    makedirs(dirname(dst))
    shutil.copyfile(src, dst)
    return dst


@pytest.fixture(scope='function')
def app(instance_path, static_folder):
    """Flask application."""
    app_ = Flask(
        'test',
        instance_path=instance_path,
        static_folder=static_folder
    )
    FlaskWebpackExt(app_)
    return app_


@pytest.yield_fixture(scope='function')
def appctx(app):
    """App in application context."""
    with app.app_context():
        yield app


@pytest.fixture(scope='function')
def ext(app):
    """Extension instance."""
    return app.extensions['flask-webpackext']


@pytest.fixture(scope='function')
def project_assets_dir(tmpdir):
    """Temporary project assets dir."""
    src = join(dirname(__file__), 'assets')
    dst = join(tmpdir, 'project')
    shutil.copytree(src, dst)
    return dst


@pytest.fixture(scope='function')
def project(app, project_assets_dir):
    """Webpack project."""
    project = WebpackProject(project_assets_dir)
    app.config.update({
        'WEBPACKEXT_PROJECT': project,
    })
    return project


@pytest.fixture(scope='function')
def projecttpl(app):
    """Webpack project."""
    project = WebpackTemplateProject(__name__, 'assets')
    app.config.update({
        'WEBPACKEXT_PROJECT': project,
    })
    return project


@pytest.fixture(scope='function')
def bundles():
    """Webpack bundles."""
    return (
        WebpackBundle(__name__, 'bundle1', entry={'app1': './app1.js'}),
        WebpackBundle(__name__, 'bundle2', entry={'app2': './app2.js'}),
    )


@pytest.fixture(scope='function')
def projectbundle(app, bundles):
    """Webpack bundle project."""
    project = WebpackBundleProject(__name__, 'assetsbundle', bundles=bundles)
    app.config.update({
        'WEBPACKEXT_PROJECT': project,
    })
    return project


@pytest.fixture(scope='function')
def cli_obj(app):
    """Script info."""
    return ScriptInfo(create_app=lambda info: app)


@pytest.fixture(scope='function')
def runner():
    """CLI Runner."""
    return CliRunner()



# -*- coding: utf-8 -*-
#
# This file is part of Flask-WebpackExt
# Copyright (C) 2017 CERN.
#
# Flask-WebpackExt is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Errors."""


class ManifestKeyNotFoundError(Exception):
    """Manifest key not found."""



# -*- coding: utf-8 -*-
#
# This file is part of Flask-WebpackExt
# Copyright (C) 2017 CERN.
#
# Flask-WebpackExt is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Python 2/3 compatibility layer."""

from __future__ import absolute_import, print_function

import sys

PY3 = sys.version_info[0] == 3

if PY3:
    string_types = str,
else:
    string_types = basestring,



# -*- coding: utf-8 -*-
#
# This file is part of Flask-WebpackExt
# Copyright (C) 2017 CERN.
#
# Flask-WebpackExt is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Version information for Flask-WebpackExt.

This file is imported by ``flask_webpackext.__init__``,
and parsed by ``setup.py``.
"""

from __future__ import absolute_import, print_function

__version__ = '1.0.1'



# -*- coding: utf-8 -*-
#
# This file is part of Flask-WebpackExt
# Copyright (C) 2017 CERN.
#
# Flask-WebpackExt is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Module tests."""

from __future__ import absolute_import, print_function

from os.path import exists, join

import pytest
from click.testing import CliRunner
from flask import Flask

from flask_webpackext import FlaskWebpackExt
from flask_webpackext.cli import webpack


def test_webpack(project):
    """Test webpack command."""
    runner = CliRunner()
    result = runner.invoke(webpack)
    assert result.exit_code == 0


@pytest.mark.parametrize('cmd', ['create', 'clean'])
def test_simple_create_clean(cmd, runner, project, cli_obj):
    """Test create/clean command on simple project."""
    result = runner.invoke(webpack, [cmd], obj=cli_obj)
    assert result.exit_code == 0
    assert 'Nothing to do' in result.output


def test_simple_chain(runner, project, cli_obj):
    """Test chaining commands."""
    result = runner.invoke(webpack, ['create', 'clean', ], obj=cli_obj)
    assert result.exit_code == 0
    assert 'Nothing to do' in result.output


def test_simple_install(project, project_assets_dir, runner, cli_obj):
    """Test install command on simple project."""
    node_dir = join(project_assets_dir, 'node_modules')
    assert not exists(node_dir)
    result = runner.invoke(webpack, ['install'], obj=cli_obj)
    assert result.exit_code == 0
    assert 'Installed webpack project.' in result.output
    assert exists(node_dir)

    # Test args passing
    result = runner.invoke(webpack, ['install', '--report'], obj=cli_obj)
    assert result.exit_code == 0


def test_simple_build(project, project_assets_dir, runner, cli_obj):
    """Test build comand on simple project."""
    bundle_js = join(project_assets_dir, 'build/bundle.js')
    assert not exists(bundle_js)
    result = runner.invoke(webpack, ['build'], obj=cli_obj)
    assert result.exit_code == 0
    assert 'Built webpack project.' in result.output
    assert exists(bundle_js)


def test_simple_run(project, project_assets_dir, runner, cli_obj):
    """Test run command."""
    bundle_js = join(project_assets_dir, 'build/bundle.js')
    assert not exists(bundle_js)
    result = runner.invoke(webpack, ['run', 'build'], obj=cli_obj)
    assert result.exit_code == 0
    assert 'Executed NPM script "build"' in result.output
    assert exists(bundle_js)


def test_simple_run_error(project, project_assets_dir, runner, cli_obj):
    """Test run command with invalid script."""
    result = runner.invoke(webpack, ['run', 'invalid'], obj=cli_obj)
    assert result.exit_code != 0


def test_simple_buildall(project, project_assets_dir, runner, cli_obj):
    """Test buildall command on simple project."""
    bundle_js = join(project_assets_dir, 'build/bundle.js')
    node_dir = join(project_assets_dir, 'node_modules')
    assert not exists(bundle_js)
    assert not exists(node_dir)
    result = runner.invoke(webpack, ['buildall'], obj=cli_obj)
    assert result.exit_code == 0
    assert exists(bundle_js)
    assert exists(node_dir)


def test_tpl_create(projecttpl, runner, cli_obj, app):
    """Test create command with template project."""
    pkgfile = join(app.instance_path, 'assets/package.json')
    assert not exists(pkgfile)
    result = runner.invoke(webpack, ['create'], obj=cli_obj)
    assert result.exit_code == 0
    assert 'Created webpack project.' in result.output
    assert exists(pkgfile)


def test_tpl_build(projecttpl, runner, cli_obj, app):
    """Test build command with template project."""
    bundle_js = join(app.instance_path, 'assets/build/bundle.js')
    assert not exists(bundle_js)
    result = runner.invoke(webpack, ['create', 'build'], obj=cli_obj)
    assert result.exit_code == 0
    assert exists(bundle_js)



# -*- coding: utf-8 -*-
#
# This file is part of Flask-WebpackExt
# Copyright (C) 2017, 2018 CERN.
#
# Flask-WebpackExt is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Webpack integration for Flask."""

from __future__ import absolute_import, print_function

WEBPACKEXT_MANIFEST_LOADER = None
"""Manifest loader use to load manifest. By default ``JinjaManifestLoader``.
"""

WEBPACKEXT_MANIFEST_PATH = 'dist/manifest.json'
"""Path to manifest file relative to static folder."""

WEBPACKEXT_PROJECT = None
"""Webpack project."""

WEBPACKEXT_PROJECT_BUILDDIR = None
"""Directory where Webpack project should be copied to prior to build. By
default ``assets``.
"""

WEBPACKEXT_PROJECT_DISTDIR = None
"""Directory where Webpack output files should be written to. By default
``dist``.
"""

WEBPACKEXT_PROJECT_DISTURL = None
"""URL path to where Webpack output files are accessible. By default ``dist``.
"""

WEBPACKEXT_STORAGE_CLS = None
"""Default storage class. By default ``FileStorage``.
"""



# -*- coding: utf-8 -*-
#
# This file is part of Flask-WebpackExt
# Copyright (C) 2018 CERN.
#
# Flask-WebpackExt is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Webpack bundle APIs for Flask-WebpackExt."""

from __future__ import absolute_import, print_function

from os.path import join

from flask.helpers import get_root_path
from pywebpack import WebpackBundle as PyWebpackBundle


class WebpackBundle(PyWebpackBundle):
    """Flask webpack bundle."""

    def __init__(self, import_name, folder, **kwargs):
        """Initialize bundle.

        :param import_name: Name of the module where the WebpackBundle class
            is instantiated. It is used to determine the absolute path to the
            ``folder`` where the assets are located.
        :param folder: Relative path to the assets.
        :param kwargs: Keyword arguments directly passed to
            :class:`pywebpack.bundle.WebpackBundle`.
        """
        super(WebpackBundle, self).__init__(
            join(get_root_path(import_name), folder),
            **kwargs
        )



# -*- coding: utf-8 -*-
#
# This file is part of Flask-WebpackExt
# Copyright (C) 2017 CERN.
#
# Flask-WebpackExt is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Module tests."""

from __future__ import absolute_import, print_function

import sys

import pytest
from flask import render_template_string
from pywebpack import ManifestLoader

from flask_webpackext.manifest import JinjaManifestLoader


def test_manifest_loading(appctx, ext, project, manifest):
    """Test manifest loading."""
    assert ext.manifest_loader == JinjaManifestLoader
    m = ext.manifest
    assert m.app
    assert m['app']


def test_manifest_rendering(appctx, ext, project, manifest):
    """Test manifest loading."""
    tpl = '{{ webpack.app }}'
    output = '<script src="/static/dist/app.js"></script>'
    assert render_template_string(tpl) == output


def test_manifest_loader_conf(app, ext, project, manifest):
    """Test manifest loading."""
    app.config.update({
        'WEBPACKEXT_MANIFEST_LOADER': 'pywebpack:ManifestLoader',
    })
    assert ext.manifest_loader == ManifestLoader


def test_manifest_nopath(app, ext):
    """Test no manifest path."""
    app.config['WEBPACKEXT_MANIFEST_PATH'] = None
    assert ext.manifest is None


def test_manifest_invalid_path(app, appctx, ext):
    """Test no manifest path."""
    app.config['WEBPACKEXT_MANIFEST_PATH'] = 'invalid/path.json'

    pytest.raises(
        FileNotFoundError if sys.version_info[0] == 3 else IOError,
        getattr, ext, 'manifest'
    )



# -*- coding: utf-8 -*-
#
# This file is part of Flask-WebpackExt
# Copyright (C) 2017, 2018 CERN.
#
# Flask-WebpackExt is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""CLI for Flask-WebpackExt."""

from __future__ import absolute_import, print_function

import click
from flask import current_app
from flask.cli import with_appcontext

from flask_webpackext import current_webpack


def _run(cmd, unavailable_msg, success_msg):
    project = current_webpack.project
    if not hasattr(project, cmd):
        click.secho(unavailable_msg, fg='yellow')
    else:
        getattr(current_webpack.project, cmd)()
        click.secho(success_msg, fg='green')


@click.group(chain=True)
@with_appcontext
def webpack():
    """Webpack commands."""


@webpack.command()
@with_appcontext
def create():
    """Create webpack project. This will fetch the assets."""
    _run('create', 'Nothing to do for project.', 'Created webpack project.')


@webpack.command()
@with_appcontext
def clean():
    """Remove created webpack project."""
    _run('clean', 'Nothing to do for project.', 'Cleaned webpack project.')


@webpack.command(context_settings={'ignore_unknown_options': True})
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
@with_appcontext
def install(args):
    """Run NPM install."""
    current_webpack.project.install(*args)
    click.secho('Installed webpack project.', fg='green')


@webpack.command(context_settings={'ignore_unknown_options': True})
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
@with_appcontext
def build(args):
    """Run NPM build-script."""
    current_webpack.project.build(*args)
    click.secho('Built webpack project.', fg='green')


@webpack.command()
@with_appcontext
def buildall():
    """Create, install and build webpack project."""
    current_webpack.project.buildall()
    click.secho('Created, installed and built webpack project.', fg='green')


@webpack.command(context_settings={'ignore_unknown_options': True})
@click.argument('script')
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
@with_appcontext
def run(script, args):
    """Run an NPM script."""
    try:
        current_webpack.project.run(script, *args)
        click.secho('Executed NPM script "{}".'.format(script), fg='green')
    except RuntimeError:
        raise click.BadParameter(
            '"{}" is not a valid NPM script.'.format(script))



# -*- coding: utf-8 -*-
#
# This file is part of Flask-WebpackExt
# Copyright (C) 2017 CERN.
#
# Flask-WebpackExt is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Module tests."""

from __future__ import absolute_import, print_function

from os.path import exists, join


def test_project_path(app, projecttpl, appctx):
    """Test project path."""
    assert projecttpl.path == app.config['WEBPACKEXT_PROJECT_BUILDDIR']


def test_project_storage_cls(ext, projecttpl, appctx):
    """Test project path."""
    assert projecttpl.storage_cls == ext.storage_cls


def test_bundle_project(projectbundle, appctx, static_folder):
    """Test bundle project."""
    out = join(static_folder, 'dist')
    p = projectbundle

    # Test create()
    assert not exists(p.path)
    p.create()
    assert exists(p.path)
    for f in ['app1.js', 'app2.js', 'webpack.config.js', 'config.json']:
        assert exists(join(p.path, f))

    # Test build()
    files = ['app1.js', 'app2.js']
    assert all([not exists(join(out, f)) for f in files])
    p.build()
    assert all([exists(join(out, f)) for f in files])



# -*- coding: utf-8 -*-
#
# This file is part of Flask-WebpackExt
# Copyright (C) 2017 CERN.
#
# Flask-WebpackExt is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Proxy to current extension."""

from __future__ import absolute_import, print_function

from flask import current_app
from werkzeug.local import LocalProxy

current_webpack = LocalProxy(
    lambda: current_app.extensions['flask-webpackext'])
"""Proxy to current extension."""

current_manifest = LocalProxy(
    lambda: current_app.extensions['flask-webpackext'].manifest)
"""Proxy to current manifest."""



# -*- coding: utf-8 -*-
#
# This file is part of Flask-WebpackExt
# Copyright (C) 2017 CERN.
#
# Flask-WebpackExt is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Module tests."""

from __future__ import absolute_import, print_function

from flask import Flask

from flask_webpackext import FlaskWebpackExt


def test_version():
    """Test version import."""
    from flask_webpackext import __version__
    assert __version__


def test_init():
    """Test extension initialization."""
    app = Flask('testapp')
    ext = FlaskWebpackExt(app)
    assert 'flask-webpackext' in app.extensions

    app = Flask('testapp')
    ext = FlaskWebpackExt()
    assert 'flask-webpackext' not in app.extensions
    ext.init_app(app)
    assert 'flask-webpackext' in app.extensions



# -*- coding: utf-8 -*-
#
# This file is part of Flask-WebpackExt
# Copyright (C) 2017, 2018 CERN.
#
# Flask-WebpackExt is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

r"""Webpack integration for Flask.

A simple example
----------------

Let's create a basic Flask application inside ``app.py``. Jinja templates will
reside in the ``templates`` folder and the static files, to be included in the
template, in the ``instance/static`` folder:

.. code-block:: python

    from flask import Flask

    app = Flask(
        __name__,
        template_folder='templates',
        static_folder='instance/static'
    )

Let's add a template and an asset to our project:

.. code-block:: console

   src/
     js/
       main.js
   templates/
     index.html
   app.py

The asset will be a simple javascript file:

.. code-block:: javascript

   import $ from 'jquery'

   $(document).ready(function(){
     $("#msg").html("I am webpacked.");
   });

The content of the ``index.html`` we will define later.

To build our asset, let's create the Webpack configuration in
``src/webpack.config.js``:

.. code-block:: javascript

    var path = require('path');
    var config = require('./config');
    var ManifestPlugin = require('webpack-manifest-plugin');

    module.exports = {
        context: config.build.context,
        entry: {
            app: "./js/main.js",
        },
        output: {
            path: config.build.assetsPath,
            filename: 'js/[name].[chunkhash].js',
            publicPath: config.build.assetsURL
        },
        plugins: [
            new ManifestPlugin({
                fileName: 'manifest.json',
                stripSrc: true,
                publicPath: config.build.assetsURL
            })
        ]
    }

You might have noticed that an import of a file which we did not create has
been added, ``config.js``. This is because this file is generated by
Flask-WebpackExt when we run the ``flask webpack create`` command and it will
contain the following information::

    {
      "build": {
          // Absolute path to the directory where Webpack outputs files
          // will be written to.
          "assetsPath": "/private/tmp/testproject/instance/static/dist",

          // URL to access the built assets.
          "assetsURL": "/static/dist/",

          // Absolute path to the generated assets directory.
          "context": "/private/tmp/testproject/instance/assets",

          // Boolean which represents if Flask debug is on.
          "debug": false,

          // Absolute path to the generated static directory.
          "staticPath": "/private/tmp/testproject/instance/static",

          // URL to access the static files.
          "staticURL": "/static/"
      }
    }


This is really important since it is Flask, and only Flask, who knows where
the application path for assets is, so through this configuration we tell
Webpack where to move the bundles.

Also, we will need a ``package.json`` for the npm dependencies, with a run
script ``build`` that executes webpack, in ``src/package.json`` and will be
triggered by ``flask webpack build``:

.. code-block:: JSON

    {
        "private": true,
        "name": "example",
        "version": "0.0.1",
        "author": "myself",
        "license": "WTFPL",
        "description": "example",
        "scripts": {
            "build": "webpack --config webpack.config.js"
        },
        "dependencies": {
            "jquery": "^3.2.1"
        },
        "devDependencies": {
            "webpack-manifest-plugin": "^2.0.4"
        }
    }


We can now define, in ``app.py`` a :class:`~project.WebpackTemplateProject` to
integrate Webpack with our Flask application and build our assets:

.. code-block:: python

    from flask_webpackext.project import WebpackTemplateProject
    from flask_webpackext import FlaskWebpackExt

    project = WebpackTemplateProject(
        __name__,
        project_folder='src',
        config_path='config.json',
    )

    app.config.update(dict(
        WEBPACKEXT_PROJECT=project,
    ))

    # Initialize extension
    FlaskWebpackExt(app)

Since Flask-WebpackExt creates a new template global function called
``webpack`` to inject the assets in templates, we can use it to include the
assets in our template ``index.html``.

.. code-block:: html

    <!DOCTYPE html>
    <html lang="en">
        <head>
            <meta charset="utf-8">
        </head>
        <body>
            <div id="msg"></div>
            {{ webpack['app.js'] }}
        </body>
    </html>

Finally, we will expose our template through a Flask view:

.. code-block:: python

    from flask import render_template

    @app.route('/')
    def index():
        return render_template('index.html')

At this point we are ready to build the assets:

.. code-block:: console

    $ flask webpack buildall

The command will copy all files from the `src` folder to the application
instance folder designated for the Webpack project, download the npm packages
and run Webpack to build our assets.

Alternatively, we can execute each build step separately. To copy all sources
to the working directory:

.. code-block:: console

    $ flask webpack create

To run `npm install` command and download all dependencies:

.. code-block:: console

    $ flask webpack build

To run `npm run build` and execute what you have defined in the
`package.json`:

.. code-block:: console

    $ flask webpack install

Now you can run the application and see if the assets are loaded:

.. code-block:: console

   $ export FLASK_APP=app.py
   $ flask run
   $ firefox http://127.0.0.1:5000/

Assets for multiple Flask modules
---------------------------------

When working with large applications, assets are usually split in different
modules. Flask-WebpackExt uses `pywebpack`_ which allows to solve this problem
by using:

- :class:`~bundle.WebpackBundle` to declare the needed assets and npm
  dependencies for each module.
- :class:`~project.WebpackBundleProject` to gather all bundles under one
  Webpack project.

An example project with two modules could be structured as follows:

.. code-block:: console

  buildconfig/
    package.json
    webpack.config.js
  modules/
    module1/
      js/
      css/
    module2/
      js/
      css/
  templates/
    index.html
  app.py

Let's start with the definition of the first bundle ``bundle1``. ``bundle2``
would be similar:

.. code-block:: python

    from flask_webpackext import WebpackBundle
    bundle1 = WebpackBundle(
        __name__,
        './modules/module1/static',
        entry={
            'module1-app': './js/module1-app.js',
        },
        dependencies={
            'jquery': '^3.2.1'
        }
    )

We now define a :class:`~project.WebpackBundleProject` to put together all the
bundles and integrate with Flask:

.. code-block:: python

    from module1 import bundle1
    from module2 import bundle2
    from flask_webpackext import WebpackBundleProject

    myproject = WebpackBundleProject(
        __name__,
        project_folder='assets',
        config_path='src/config.json',
        bundles=[bundle1, bundle2],
    )

    app.config.update(dict(
        WEBPACKEXT_PROJECT=myproject,
    ))

    # ...

At this point, we can create the application and ``config.json`` will be
generated. The main difference with the previous ``config.json`` is that
Flask-WebpackExt now knows about the bundles, and it will add them under the
key ``entry``::

    {
        "build": {...},
        "entry": {
            "module1": "./js/module1.js",
            "module2": "./js/module2.js"
        }
    }

Because entries are now included in the ``config.json``, we can dynamically
load them into ``webpack.config.js``::

    var config = require('./config')

    module.exports = {
        context: config.build.context,
        entry: config.entry,
        ...
    };


Manifest
--------

Flask-WebpackExt can load the ``manifest.json`` created by webpack when
building the assets. You have to add the plugin in the ``webpack.config.js``
to generate it::

    ...
    var ManifestPlugin = require('webpack-manifest-plugin');

    module.exports = {
        ...
        plugins: [
            new ManifestPlugin({
                fileName: 'manifest.json',
                stripSrc: true,
                publicPath: config.build.assetsURL
            })
        ]
    };

The generated ``manifest.json`` will look like :

.. code-block:: javascript

    {
    "module1.js": "/static/dist/js/module1.4adb22699eb1a5698794.js",
    "module2.js": "/static/dist/js/module2.85e58794420201dc1426.js"
    }

By default, Flask-WebpackExt will look for the manifest file in
`WEBPACKEXT_MANIFEST_PATH` config variable, and if it exists, it will load the
file and make it available inside Jinja templates.

The injected asset in the generated html template will look similar to this::

    <script src="/static/dist/mymodule-app.8817a9d4faccbc712aa7.js"></script>

You can read more about it on `pywebpack`_ documentation.

.. _pywebpack: https://pywebpack.readthedocs.io
"""

from __future__ import absolute_import, print_function

from pywebpack import WebpackProject

from .bundle import WebpackBundle
from .ext import FlaskWebpackExt
from .project import WebpackBundleProject, WebpackTemplateProject
from .proxies import current_webpack, current_manifest
from .version import __version__

__all__ = (
    '__version__',
    'current_manifest',
    'current_webpack',
    'FlaskWebpackExt',
    'WebpackBundle',
    'WebpackBundleProject',
    'WebpackProject',
    'WebpackTemplateProject',
)
